/**
 * Developed by: The Open Systems Lab
 *             University of Illinois at Urbana-Champaign
 *             Department of Computer Science
 *             Urbana, IL 61801
 *             http://osl.cs.uiuc.edu
 *
 * Contact: http://osl.cs.uiuc.edu/soter
 *
 * Copyright (c) 1998-2011
 * The University of Illinois Board of Trustees.
 *    All Rights Reserved.
 * 
 * Distributed under license: http://osl.cs.uiuc.edu/soter/LICENSE
 * 
 */
package edu.illinois.soter.analysis;

import java.util.HashSet;
import java.util.Set;

import com.ibm.wala.classLoader.CallSiteReference;
import com.ibm.wala.classLoader.IClass;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.ipa.callgraph.impl.AbstractRootMethod;
import com.ibm.wala.ipa.callgraph.impl.DefaultEntrypoint;
import com.ibm.wala.ipa.cha.IClassHierarchy;
import com.ibm.wala.shrikeBT.IInvokeInstruction;
import com.ibm.wala.ssa.SSANewInstruction;
import com.ibm.wala.types.TypeReference;

/**
 * 
 * Entrypoint that ensures *complete* initialization of the receiver and
 * parameters (i.e. all their fields are (transitively) initialized as well)
 * 
 * @author Stas Negara
 * 
 */
public class FieldSensitiveEntrypoint extends DefaultEntrypoint {

	private final IClassHierarchy classHierachy;

	private AbstractRootMethod rootMethod;

	//Classes that can be uninitialized
	private static final Set<String> uninitializableClasses = new HashSet<String>(11);

	static {
		uninitializableClasses.add("Ljava/lang/Boolean");
		uninitializableClasses.add("Ljava/lang/Character");
		uninitializableClasses.add("Ljava/lang/String");
		uninitializableClasses.add("Ljava/lang/StringBuilder");
		uninitializableClasses.add("Ljava/lang/StringBuffer");
		uninitializableClasses.add("Ljava/lang/Byte");
		uninitializableClasses.add("Ljava/lang/Short");
		uninitializableClasses.add("Ljava/lang/Integer");
		uninitializableClasses.add("Ljava/lang/Long");
		uninitializableClasses.add("Ljava/lang/Float");
		uninitializableClasses.add("Ljava/lang/Double");
	}

	public FieldSensitiveEntrypoint(IMethod method, IClassHierarchy classHeirarchy) {
		super(method, classHeirarchy);
		this.classHierachy = classHeirarchy;
	}

	@Override
	protected int makeArgument(AbstractRootMethod rootMethod, int argumentIndex) {
		this.rootMethod = rootMethod;
		TypeReference[] typeReferences = getParameterTypes(argumentIndex);
		//Although WALA supports an array, in the current implementation it contains a single element
		if (typeReferences.length == 0 || typeReferences.length > 1) {
			throw new RuntimeException("None or multiple types for argumentIndex " + argumentIndex + " in entry point: " + this);
		}
		return initializeArgument(new HashSet<IMethod>(), typeReferences[0]);
	}

	/**
	 * TODO: Could be optimized (i.e. avoid infinite recursion but without the
	 * need to create new sets for every branch)?
	 * 
	 * @param appliedInitMethods
	 *            Used to keep track of initializations that were already
	 *            applied in order to avoid infinite recursion when a
	 *            constructor of a class accepts an instance of the same class
	 *            (e.g. java.lang.String(java.lang.String) ).
	 * @param typeReference
	 * @return
	 */
	private int initializeArgument(Set<IMethod> appliedInitMethods, TypeReference typeReference) {
		int thisValueNumber;
		if (typeReference.isPrimitiveType()) {
			thisValueNumber = rootMethod.addLocal();
		} else {
			SSANewInstruction newInstruction = rootMethod.addAllocation(typeReference);
			if (newInstruction == null) {
				throw new RuntimeException("Failed to allocate type: " + typeReference);
			}
			thisValueNumber = newInstruction.getDef();
			IClass klass = classHierachy.lookupClass(typeReference);
			if (!uninitializableClasses.contains(klass.getName().toString())) {
				for (IMethod method : klass.getDeclaredMethods()) {
					if (AnalysisHelper.isInitMethod(method) && !appliedInitMethods.contains(method)) {
						appliedInitMethods.add(method);
						int numberOfParameters = method.getNumberOfParameters();
						int parameterValues[] = new int[numberOfParameters];
						parameterValues[0] = thisValueNumber;
						for (int i = 1; i < numberOfParameters; i++) {
							parameterValues[i] = initializeArgument(new HashSet<IMethod>(appliedInitMethods), method.getParameterType(i));
						}
						int nextProgramCounter = rootMethod.getStatements().length;
						rootMethod.addInvocation(parameterValues, CallSiteReference.make(nextProgramCounter, method.getReference(), IInvokeInstruction.Dispatch.SPECIAL));
					}
				}
			}
		}
		return thisValueNumber;
	}

}
