/**
 * Developed by: The Open Systems Lab
 *             University of Illinois at Urbana-Champaign
 *             Department of Computer Science
 *             Urbana, IL 61801
 *             http://osl.cs.uiuc.edu
 *
 * Contact: http://osl.cs.uiuc.edu/soter
 *
 * Copyright (c) 1998-2011
 * The University of Illinois Board of Trustees.
 *    All Rights Reserved.
 * 
 * Distributed under license: http://osl.cs.uiuc.edu/soter/LICENSE
 * 
 */
package edu.illinois.soter.analysis;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.NullProgressMonitor;

import com.ibm.wala.analysis.pointers.BasicHeapGraph;
import com.ibm.wala.classLoader.CallSiteReference;
import com.ibm.wala.classLoader.IClass;
import com.ibm.wala.classLoader.IField;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.core.tests.callGraph.CallGraphTestUtil;
import com.ibm.wala.dataflow.graph.BitVectorFramework;
import com.ibm.wala.dataflow.graph.BitVectorSolver;
import com.ibm.wala.fixpoint.BitVectorVariable;
import com.ibm.wala.ipa.callgraph.AnalysisCache;
import com.ibm.wala.ipa.callgraph.AnalysisOptions;
import com.ibm.wala.ipa.callgraph.AnalysisScope;
import com.ibm.wala.ipa.callgraph.CGNode;
import com.ibm.wala.ipa.callgraph.CallGraph;
import com.ibm.wala.ipa.callgraph.CallGraphBuilder;
import com.ibm.wala.ipa.callgraph.ClassTargetSelector;
import com.ibm.wala.ipa.callgraph.Entrypoint;
import com.ibm.wala.ipa.callgraph.MethodTargetSelector;
import com.ibm.wala.ipa.callgraph.impl.Util;
import com.ibm.wala.ipa.callgraph.propagation.HeapModel;
import com.ibm.wala.ipa.callgraph.propagation.InstanceFieldKey;
import com.ibm.wala.ipa.callgraph.propagation.InstanceKey;
import com.ibm.wala.ipa.callgraph.propagation.PointerAnalysis;
import com.ibm.wala.ipa.callgraph.propagation.PointerKey;
import com.ibm.wala.ipa.callgraph.propagation.SSAPropagationCallGraphBuilder;
import com.ibm.wala.ipa.callgraph.propagation.cfa.ZeroXCFABuilder;
import com.ibm.wala.ipa.callgraph.propagation.cfa.ZeroXInstanceKeys;
import com.ibm.wala.ipa.cha.ClassHierarchy;
import com.ibm.wala.ipa.summaries.BypassClassTargetSelector;
import com.ibm.wala.ipa.summaries.BypassMethodTargetSelector;
import com.ibm.wala.ipa.summaries.XMLMethodSummaryReader;
import com.ibm.wala.ssa.IR;
import com.ibm.wala.ssa.ISSABasicBlock;
import com.ibm.wala.ssa.SSAAbstractInvokeInstruction;
import com.ibm.wala.ssa.SSACFG;
import com.ibm.wala.ssa.SSAInstruction;
import com.ibm.wala.types.ClassLoaderReference;
import com.ibm.wala.types.MethodReference;
import com.ibm.wala.types.Selector;
import com.ibm.wala.types.TypeReference;
import com.ibm.wala.util.CancelException;
import com.ibm.wala.util.collections.HashSetFactory;
import com.ibm.wala.util.config.AnalysisScopeReader;
import com.ibm.wala.util.graph.impl.GraphInverter;
import com.ibm.wala.util.intset.BitVector;
import com.ibm.wala.util.intset.IntIterator;
import com.ibm.wala.util.intset.IntSet;
import com.ibm.wala.util.intset.MutableMapping;
import com.ibm.wala.util.intset.OrdinalSetMapping;
import com.ibm.wala.util.io.FileProvider;
import com.ibm.wala.util.strings.Atom;
import com.ibm.wala.util.warnings.WalaException;

import edu.illinois.soter.Debugger;
import edu.illinois.soter.analysis.contextselectors.ReceiverInstanceContextSelector;
import edu.illinois.soter.analysis.transferfunctionproviders.GlobalLiveVariablesTransferFunctionProvider;
import edu.illinois.soter.analysis.transferfunctionproviders.LocalLiveVariablesTransferFunctionProvider;
import edu.illinois.soter.messagedata.MessageInvocation;

/**
 * 
 * @author Stas Negara
 * 
 */
public abstract class OwnershipTransferAnalysis {

	//TODO: Is this flag really needed, considering that static fields are global anyway (i.e. can be accessed by any actor anyway)?
	//(private static fields are not global, they can be accessed only by "this", but at the same time, they are pointed by "this",
	//while "this" is always live, which makes the tool consider such private static fields anyway).
	public static final boolean IS_IGNORING_STATIC_FIELDS = Boolean.TRUE;

	private final String basicActorTypeName;
	private final String classPath;

	protected TypeReference basicActorReference;
	protected AnalysisScope analysisScope;
	protected AnalysisOptions analysisOptions;
	protected ClassHierarchy classHierarchy;
	protected CallGraph callGraph;
	protected PointerAnalysis pointerAnalysis;
	protected HeapModel heapModel;
	protected BasicHeapGraph basicHeapGraph;

	private final Set<Object> itemsPointedByStaticFields = new HashSet<Object>();

	protected final Map<CGNode, Map<CallSiteReference, MessageInvocation>> messageInvocations = new HashMap<CGNode, Map<CallSiteReference, MessageInvocation>>();

	public OwnershipTransferAnalysis(String basicActorTypeName, String classPath) {
		this.basicActorTypeName = basicActorTypeName;
		this.classPath = classPath;
	}

	public Map<CGNode, Map<CallSiteReference, MessageInvocation>> perform() throws WalaException, IllegalArgumentException, CancelException, IOException {
		initAnalysis();
		performAnalysis();
		AnalysisHelper.displayGraph(callGraph);
		//AnalysisHelper.displayGraph(basicHeapGraph);
		return messageInvocations;
	}

	private void initAnalysis() throws WalaException, IllegalArgumentException, CancelException, IOException {
		analysisScope = AnalysisScopeReader.makeJavaBinaryAnalysisScope(classPath, FileProvider.getInputStream(CallGraphTestUtil.REGRESSION_EXCLUSIONS));
		classHierarchy = ClassHierarchy.make(analysisScope);
		basicActorReference = getReferenceForTypeName(basicActorTypeName);

		analysisOptions = new AnalysisOptions(analysisScope, makeCallGraphEntrypoints());

		CallGraphBuilder builder = makeZeroOneCFABuilder();

		callGraph = builder.makeCallGraph(analysisOptions, null);
		pointerAnalysis = builder.getPointerAnalysis();
		heapModel = pointerAnalysis.getHeapModel();
		basicHeapGraph = new BasicHeapGraph(pointerAnalysis, callGraph);

		setupIgnoredFields();

		populateItemsPointedByStaticFields();
	}

	private void populateItemsPointedByStaticFields() {
		if (!IS_IGNORING_STATIC_FIELDS) {
			for (IField staticField : getAllStaticFields()) {
				PointerKey staticFieldPointer = heapModel.getPointerKeyForStaticField(staticField);
				Set<Object> pointedItems = computeTransitiveClosureOfPointedItems(staticFieldPointer);
				itemsPointedByStaticFields.addAll(pointedItems);
			}
		}
	}

	private Set<IField> getAllStaticFields() {
		Set<IField> staticFields = new HashSet<IField>();
		Iterator<IClass> classIterator = classHierarchy.iterator();
		while (classIterator.hasNext()) {
			IClass klass = classIterator.next();
			if (!AnalysisHelper.isJDKClass(klass)) {
				staticFields.addAll(klass.getAllStaticFields());
			}
		}
		return staticFields;
	}

	protected final TypeReference getReferenceForTypeName(String typeName) {
		ClassLoaderReference applicationLoaderReference = analysisScope.getApplicationLoader();
		return TypeReference.findOrCreate(applicationLoaderReference, typeName);
	}

	private SSAPropagationCallGraphBuilder makeZeroOneCFABuilder() {
		if (analysisOptions == null) {
			throw new IllegalArgumentException("analysisOptions is null");
		}
		Util.addDefaultSelectors(analysisOptions, classHierarchy);
		addCustomBypassLogic();
		return ZeroXCFABuilder.make(classHierarchy, analysisOptions, new AnalysisCache(), new ReceiverInstanceContextSelector(), null, ZeroXInstanceKeys.ALLOCATIONS |
		/* ZeroXInstanceKeys.SMUSH_MANY | ZeroXInstanceKeys.SMUSH_STRINGS | */ZeroXInstanceKeys.SMUSH_THROWABLES);
	}

	private void addCustomBypassLogic() throws IllegalArgumentException {
		ClassLoader classLoader = Util.class.getClassLoader();
		if (classLoader == null) {
			throw new IllegalArgumentException("classLoader is null");
		}

		InputStream inputStream = classLoader.getResourceAsStream(Util.nativeSpec);
		XMLMethodSummaryReader summary = new XMLMethodSummaryReader(inputStream, analysisScope);

		MethodTargetSelector customMethodTargetSelector = getCustomBypassMethodTargetSelector(summary);
		analysisOptions.setSelector(customMethodTargetSelector);

		ClassTargetSelector classTargetSelector = new BypassClassTargetSelector(analysisOptions.getClassTargetSelector(), summary.getAllocatableClasses(), classHierarchy,
				classHierarchy.getLoader(analysisScope.getLoader(Atom.findOrCreateUnicodeAtom("Synthetic"))));
		analysisOptions.setSelector(classTargetSelector);
	}

	protected BypassMethodTargetSelector getCustomBypassMethodTargetSelector(XMLMethodSummaryReader summary) {
		return new BypassMethodTargetSelector(analysisOptions.getMethodTargetSelector(), summary.getSummaries(), summary.getIgnoredPackages(), classHierarchy);
	}

	protected abstract void setupIgnoredFields();

	protected abstract boolean isIgnoredField(IField field);

	private Iterable<Entrypoint> makeCallGraphEntrypoints() {
		if (analysisScope == null) {
			throw new RuntimeException("analysisScope is null");
		}
		if (classHierarchy == null) {
			throw new RuntimeException("classHierarchy is null");
		}
		if (basicActorReference == null) {
			throw new RuntimeException("basicActorReference is null");
		}

		final Set<Entrypoint> entrypoints = HashSetFactory.make();
		populateEntrypoints(entrypoints);

		return new Iterable<Entrypoint>() {
			public Iterator<Entrypoint> iterator() {
				return entrypoints.iterator();
			}
		};
	}

	protected abstract void populateEntrypoints(Set<Entrypoint> entrypoints);

	/**
	 * For ActorFoundry analysis, should be applied before Kilim weaving.
	 * Otherwise the added kilim/Fiber; parameter points to much more than
	 * needed (i.e. everything becomes always live).
	 */
	private void performAnalysis() {
		Debugger.debugEntryPoints(callGraph);
		populateMessageInvocations();
		Set<CGNode> sendingCGNodes = messageInvocations.keySet();
		Set<CGNode> relevantCGNodes = computeTransitiveClosureOfCallers(sendingCGNodes);

		OrdinalSetMapping<PointerKey> globalValues = MutableMapping.make();
		Map<CGNode, Map<CallSiteReference, BitVector>> relevantCGNodesLocalLiveVariables = getRelevantCGNodesLocalLiveVariables(relevantCGNodes, globalValues);
		BitVectorSolver<CGNode> bitVectorSolver = solveGlobalLiveVariablesProblem(relevantCGNodes, relevantCGNodesLocalLiveVariables, globalValues);

		markPassByValueMethodInvocationsArguments(bitVectorSolver, relevantCGNodesLocalLiveVariables, globalValues);
	}

	/**
	 * For every call graph node detects all message passing call sites and
	 * stores information about them.
	 * 
	 * @return
	 */
	private void populateMessageInvocations() {
		CGNode fakeRootNode = callGraph.getFakeRootNode();
		Iterator<CGNode> iterator = callGraph.iterator();
		while (iterator.hasNext()) {
			CGNode cgNode = iterator.next();
			IR ir = cgNode.getIR();
			if (!fakeRootNode.equals(cgNode) && ir != null && isNodeConsideredForMessageInvocations(cgNode)) {
				Iterator<CallSiteReference> callSitesIterator = ir.iterateCallSites();
				Map<CallSiteReference, MessageInvocation> callSiteReferences = messageInvocations.get(cgNode);
				while (callSitesIterator.hasNext()) {
					CallSiteReference callSiteReference = callSitesIterator.next();
					int argumentsIndex = getMessageArgumentsIndex(callSiteReference.getDeclaredTarget());
					if (argumentsIndex != -1) { //found a message passing site
						MessageInvocation messageInvocation = getMessageInvocation(cgNode, callSiteReference, argumentsIndex);
						if (callSiteReferences == null) {
							callSiteReferences = new HashMap<CallSiteReference, MessageInvocation>();
							messageInvocations.put(cgNode, callSiteReferences);
						}
						callSiteReferences.put(callSiteReference, messageInvocation);
					}
				}
			}
		}
	}

	protected abstract boolean isNodeConsideredForMessageInvocations(CGNode cgNode);

	public boolean canTransform() {
		return false;
	}

	/**
	 * 
	 * @param methodReference
	 * @return The index of the argument that represents object(s) passed to an
	 *         actor in a message. If this is not an actor message passing
	 *         signature, then returns -1.
	 */
	protected abstract int getMessageArgumentsIndex(MethodReference methodReference);

	private MessageInvocation getMessageInvocation(CGNode cgNode, CallSiteReference callSiteReference, int argumentsIndex) {
		IR ir = cgNode.getIR();
		int messagePassingIndex = AnalysisHelper.getBytecodeIndexForCallSite(ir, callSiteReference);
		SSAInstruction[] instructions = ir.getInstructions();
		SSAAbstractInvokeInstruction messagePassingInstruction = (SSAAbstractInvokeInstruction) instructions[messagePassingIndex];
		int passedPointerValueNumber = messagePassingInstruction.getUse(argumentsIndex);
		MessageInvocation messageInvocation = new MessageInvocation(cgNode, callSiteReference);
		addMessageInvocationArguments(messageInvocation, messagePassingIndex, passedPointerValueNumber);
		return messageInvocation;
	}

	protected abstract void addMessageInvocationArguments(MessageInvocation messageInvocation, int messagePassingIndex, int passedPointerValueNumber);

	private Set<CGNode> computeTransitiveClosureOfCallers(Set<CGNode> initialSet) {
		Set<CGNode> transitiveClosure = new HashSet<CGNode>();
		LinkedList<CGNode> workList = new LinkedList<CGNode>(initialSet);
		while (!workList.isEmpty()) {
			CGNode cgNode = workList.poll();
			if (!transitiveClosure.contains(cgNode)) {
				transitiveClosure.add(cgNode);
				Iterator<CGNode> callersIterator = callGraph.getPredNodes(cgNode);
				while (callersIterator.hasNext()) {
					CGNode callerNode = callersIterator.next();
					workList.add(callerNode);
				}
			}
		}
		return transitiveClosure;
	}

	protected final Set<Object> computeTransitiveClosureOfPointedItems(PointerKey pointer) {
		Set<Object> allPointedItems = new HashSet<Object>();
		LinkedList<Iterator<Object>> workList = new LinkedList<Iterator<Object>>();
		workList.add(getImmediatelyPointedItemsIterator(pointer));
		while (!workList.isEmpty()) {
			Iterator<Object> pointedItemsIterator = workList.poll();
			while (pointedItemsIterator.hasNext()) {
				Object pointedItem = pointedItemsIterator.next();
				if (!allPointedItems.contains(pointedItem)) {
					if (pointedItem instanceof InstanceFieldKey && isIgnoredField(((InstanceFieldKey) pointedItem).getField())) {
						continue; //skip the instance field, which should be avoided
					}
					if (pointedItem instanceof InstanceKey && AnalysisHelper.isImmutableInstanceKey((InstanceKey) pointedItem)) {
						continue; //skip instances that are immutable (actually, this is disabled here, so should never reach this statement)
					}

					if (pointedItem instanceof InstanceKey) {
						allPointedItems.add(pointedItem);
					}

					workList.add(getImmediatelyPointedItemsIterator(pointedItem));
				}
			}
		}
		return allPointedItems;
	}

	protected final Iterator<Object> getImmediatelyPointedItemsIterator(Object object) {
		return basicHeapGraph.getSuccNodes(object);
	}

	protected final IMethod lookupMethod(String typeName, Selector methodSelector) {
		TypeReference typeReference = getReferenceForTypeName(typeName);
		IClass klass = classHierarchy.lookupClass(typeReference);
		if (klass == null) {
			throw new RuntimeException("Could not find class \"" + typeName + "\"");
		}
		return lookupMethod(klass, methodSelector);
	}

	protected final IMethod lookupMethod(IClass klass, Selector methodSelector) {
		IMethod method = klass.getMethod(methodSelector);
		if (method == null) {
			throw new RuntimeException("Could not find method \"" + methodSelector + "\" in class \"" + klass + "\"");
		}
		return method;
	}

	private Map<CallSiteReference, Set<PointerKey>> getLocalLiveVariablesForCallSites(CGNode cgNode, Set<CallSiteReference> callSites) {
		OrdinalSetMapping<PointerKey> localValues = MutableMapping.make();
		BitVectorSolver<ISSABasicBlock> bitVectorSolver = solveLocalLiveVariablesProblem(cgNode, localValues);
		Map<CallSiteReference, Set<PointerKey>> liveVariablesForCallSites = new HashMap<CallSiteReference, Set<PointerKey>>();
		for (CallSiteReference callSiteReference : callSites) {
			ISSABasicBlock[] basicBlocksForCall = cgNode.getIR().getBasicBlocksForCall(callSiteReference);
			if (basicBlocksForCall.length == 0 || basicBlocksForCall.length > 1) {
				throw new RuntimeException("None or more than one basic block for call site \"" + callSiteReference + "\"");
			}
			Set<PointerKey> pointerKeys = new HashSet<PointerKey>();
			IntSet intSet = bitVectorSolver.getIn(basicBlocksForCall[0]).getValue();
			if (intSet != null) {
				IntIterator intSetIterator = intSet.intIterator();
				while (intSetIterator.hasNext()) {
					pointerKeys.add(localValues.getMappedObject(intSetIterator.next()));
				}
			}
			liveVariablesForCallSites.put(callSiteReference, pointerKeys);
		}
		return liveVariablesForCallSites;
	}

	private BitVectorSolver<ISSABasicBlock> solveLocalLiveVariablesProblem(CGNode cgNode, OrdinalSetMapping<PointerKey> values) {
		SSACFG controlFlowGraph = cgNode.getIR().getControlFlowGraph();
		LocalLiveVariablesTransferFunctionProvider transferFunctions = new LocalLiveVariablesTransferFunctionProvider(heapModel, cgNode, values);
		//Invert the control flow graph, because live variables analysis is a *backwards* dataflow problem.
		BitVectorFramework<ISSABasicBlock, PointerKey> bitVectorFramework = new BitVectorFramework<ISSABasicBlock, PointerKey>(GraphInverter.invert(controlFlowGraph), transferFunctions, values);
		BitVectorSolver<ISSABasicBlock> bitVectorSolver = new BitVectorSolver<ISSABasicBlock>(bitVectorFramework);
		try {
			bitVectorSolver.solve(new NullProgressMonitor());
		} catch (CancelException ex) {
			ex.printStackTrace();
		}
		return bitVectorSolver;
	}

	private Set<CallSiteReference> getRelevantCallSitesForNode(CGNode cgNode, Set<CGNode> relevantCGNodes) {
		Set<CallSiteReference> relevantCallSites = new HashSet<CallSiteReference>();
		Map<CallSiteReference, MessageInvocation> sendingCallSites = messageInvocations.get(cgNode); //may be null
		if (sendingCallSites != null) {
			relevantCallSites.addAll(sendingCallSites.keySet());
		}
		Iterator<CallSiteReference> callSitesIterator = cgNode.iterateCallSites();
		while (callSitesIterator.hasNext()) {
			CallSiteReference callSiteReference = callSitesIterator.next();
			Set<CGNode> cgSiteNodes = callGraph.getPossibleTargets(cgNode, callSiteReference);
			for (CGNode cgSiteNode : cgSiteNodes) {
				if (relevantCGNodes.contains(cgSiteNode)) {
					relevantCallSites.add(callSiteReference);
					break;
				}
			}
		}
		return relevantCallSites;
	}

	private Map<CGNode, Map<CallSiteReference, BitVector>> getRelevantCGNodesLocalLiveVariables(Set<CGNode> relevantCGNodes, OrdinalSetMapping<PointerKey> globalValues) {
		Map<CGNode, Map<CallSiteReference, BitVector>> relevantCGNodesLocalLiveVariables = new HashMap<CGNode, Map<CallSiteReference, BitVector>>();
		for (CGNode cgNode : relevantCGNodes) {
			Set<CallSiteReference> relevantCallSites = getRelevantCallSitesForNode(cgNode, relevantCGNodes);
			Map<CallSiteReference, Set<PointerKey>> liveVariablesForCallSites = getLocalLiveVariablesForCallSites(cgNode, relevantCallSites);
			Map<CallSiteReference, BitVector> cgNodeLiveVariables = new HashMap<CallSiteReference, BitVector>();
			relevantCGNodesLocalLiveVariables.put(cgNode, cgNodeLiveVariables);
			Iterator<CallSiteReference> callSitesIterator = liveVariablesForCallSites.keySet().iterator();
			while (callSitesIterator.hasNext()) {
				CallSiteReference callSiteReference = callSitesIterator.next();
				BitVector bitVector = new BitVector();
				Iterator<PointerKey> pointerKeysIterator = liveVariablesForCallSites.get(callSiteReference).iterator();
				while (pointerKeysIterator.hasNext()) {
					bitVector.set(globalValues.add(pointerKeysIterator.next()));
				}
				//Ensure that "this" is always live
				//(live variables analysis does not consider "this" live, if in cgNode there are no field accesses or calls
				//after this call site)
				if (!cgNode.getMethod().isStatic()) {
					bitVector.set(globalValues.add(heapModel.getPointerKeyForLocal(cgNode, 1)));
				}
				cgNodeLiveVariables.put(callSiteReference, bitVector);
			}
		}
		return relevantCGNodesLocalLiveVariables;
	}

	private BitVectorSolver<CGNode> solveGlobalLiveVariablesProblem(Set<CGNode> relevantCGNodes, Map<CGNode, Map<CallSiteReference, BitVector>> relevantCGNodesLocalLiveVariables,
			OrdinalSetMapping<PointerKey> globalValues) {

		GlobalLiveVariablesTransferFunctionProvider transferFunctions = new GlobalLiveVariablesTransferFunctionProvider(callGraph, relevantCGNodes, relevantCGNodesLocalLiveVariables);

		BitVectorFramework<CGNode, PointerKey> bitVectorFramework = new BitVectorFramework<CGNode, PointerKey>(callGraph, transferFunctions, globalValues);

		BitVectorSolver<CGNode> bitVectorSolver = new BitVectorSolver<CGNode>(bitVectorFramework);
		try {
			bitVectorSolver.solve(new NullProgressMonitor());
		} catch (CancelException ex) {
			ex.printStackTrace();
		}
		return bitVectorSolver;
	}

	private Set<Object> getNodeLiveItems(CGNode cgNode, BitVectorSolver<CGNode> bitVectorSolver, OrdinalSetMapping<PointerKey> globalValues) {
		Set<Object> nodeLiveItems = new HashSet<Object>();
		//Use getIn, because getOut does not work as there are no node transfer functions
		IntSet value = bitVectorSolver.getIn(cgNode).getValue();
		if (value != null) {
			IntIterator intIterator = value.intIterator();
			while (intIterator.hasNext()) {
				PointerKey pointerKey = globalValues.getMappedObject(intIterator.next());
				nodeLiveItems.addAll(computeTransitiveClosureOfPointedItems(pointerKey));
			}
		}
		return nodeLiveItems;
	}

	private void markPassByValueMethodInvocationsArguments(BitVectorSolver<CGNode> bitVectorSolver, Map<CGNode, Map<CallSiteReference, BitVector>> relevantCGNodesLocalLiveVariables,
			OrdinalSetMapping<PointerKey> globalValues) {
		Set<CGNode> sendingCGNodes = messageInvocations.keySet();
		for (CGNode cgNode : sendingCGNodes) {
			Debugger.debugSendingNode(cgNode);
			Set<Object> nodeLiveItems = getNodeLiveItems(cgNode, bitVectorSolver, globalValues);
			Map<CallSiteReference, MessageInvocation> sendingCallSites = messageInvocations.get(cgNode);
			Map<CallSiteReference, BitVector> cgNodeLiveVariables = relevantCGNodesLocalLiveVariables.get(cgNode);
			Iterator<CallSiteReference> callSitesIterator = sendingCallSites.keySet().iterator();
			while (callSitesIterator.hasNext()) {
				CallSiteReference callSiteReference = callSitesIterator.next();
				BitVector bitVector = cgNodeLiveVariables.get(callSiteReference);
				BitVectorVariable variable = new BitVectorVariable();
				variable.addAll(bitVector);
				Set<Object> callSiteLiveItems = new HashSet<Object>();
				IntIterator intIterator = variable.getValue().intIterator();
				while (intIterator.hasNext()) {
					PointerKey pointerKey = globalValues.getMappedObject(intIterator.next());
					callSiteLiveItems.addAll(computeTransitiveClosureOfPointedItems(pointerKey));
				}
				callSiteLiveItems.addAll(nodeLiveItems);
				callSiteLiveItems.addAll(itemsPointedByStaticFields);
				MessageInvocation messageInvocation = sendingCallSites.get(callSiteReference);
				messageInvocation.check(callSiteLiveItems);
			}
		}
	}

}