/**
 * Developed by: The Open Systems Lab
 *             University of Illinois at Urbana-Champaign
 *             Department of Computer Science
 *             Urbana, IL 61801
 *             http://osl.cs.uiuc.edu
 *
 * Contact: http://osl.cs.uiuc.edu/soter
 *
 * Copyright (c) 1998-2011
 * The University of Illinois Board of Trustees.
 *    All Rights Reserved.
 * 
 * Distributed under license: http://osl.cs.uiuc.edu/soter/LICENSE
 * 
 */
package edu.illinois.soter.analysis.actorfoundry;

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;

import com.ibm.wala.classLoader.IClass;
import com.ibm.wala.classLoader.IField;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.classLoader.ShrikeCTMethod;
import com.ibm.wala.ipa.callgraph.CGNode;
import com.ibm.wala.ipa.callgraph.Entrypoint;
import com.ibm.wala.ipa.callgraph.propagation.PointerKey;
import com.ibm.wala.ssa.SSAArrayStoreInstruction;
import com.ibm.wala.ssa.SSAInstruction;
import com.ibm.wala.ssa.SSANewInstruction;
import com.ibm.wala.types.MethodReference;
import com.ibm.wala.types.annotations.Annotation;
import com.ibm.wala.util.strings.Atom;

import edu.illinois.soter.Debugger;
import edu.illinois.soter.analysis.OwnershipTransferAnalysis;
import edu.illinois.soter.messagedata.MessageInvocation;

/**
 * 
 * @author Stas Negara
 * 
 */
public class ActorFoundryAnalysis extends OwnershipTransferAnalysis {

	private static final String MESSAGE_ANNOTATION = "message";

	private static final String SEND_SIGNATURE_SUFFIX = ".send(Losl/manager/ActorName;Ljava/lang/String;[Ljava/io/Serializable;)V";
	private static final String CALL_SIGNATURE_SUFFIX = ".call(Losl/manager/ActorName;Ljava/lang/String;[Ljava/io/Serializable;)Ljava/lang/Object;";
	private static final String CREATE_SIGNATURE_SUFFIX_1 = ".create(Ljava/lang/Class;[Ljava/io/Serializable;)Losl/manager/ActorName;";
	private static final String CREATE_SIGNATURE_SUFFIX_2 = ".create(Losl/manager/ActorManagerName;Ljava/lang/Class;[Ljava/io/Serializable;)Losl/manager/ActorName;";

	private Collection<IClass> actorSubclasses;

	private IField lastRequestField; //Field of osl.manager.Actor class, which should be avoided in pointer analysis

	public ActorFoundryAnalysis(String classPath) {
		super("Losl/manager/Actor", classPath);
	}

	@Override
	protected void setupIgnoredFields() {
		IClass actorClass = classHierarchy.lookupClass(basicActorReference);
		lastRequestField = actorClass.getField(Atom.findOrCreateAsciiAtom("lastRequest"));
	}

	@Override
	protected boolean isIgnoredField(IField field) {
		return lastRequestField.equals(field);
	}

	@Override
	protected void populateEntrypoints(final Set<Entrypoint> entrypoints) {
		actorSubclasses = classHierarchy.computeSubClasses(basicActorReference);
		for (IClass klass : actorSubclasses) {
			//Do not consider abstract classes and Actor class itself
			if (!klass.isAbstract() && !basicActorReference.equals(klass.getReference())) {
				Iterator<IMethod> iterator = klass.getAllMethods().iterator();
				while (iterator.hasNext()) {
					ShrikeCTMethod method = (ShrikeCTMethod) iterator.next();//Should be ShrikeCTMethod
					try {
						//Annotation @message is runtime visible
						Collection<Annotation> annotations = method.getRuntimeVisibleAnnotations();
						for (Annotation annotation : annotations) {
							if (MESSAGE_ANNOTATION.equals(annotation.getType().getName().getClassName().toString())) {
								Debugger.debugEntrypointMethod(method);
								SingleInstanceEntrypoint singleInstanceEntrypoint = new SingleInstanceEntrypoint(klass, method, classHierarchy);
								entrypoints.add(singleInstanceEntrypoint);
								break;
							}
						}
					} catch (Exception ex) {
						ex.printStackTrace();
					}
				}
			}
		}
	}

	@Override
	protected boolean isNodeConsideredForMessageInvocations(CGNode cgNode) {
		//Consider only classes derived from osl.manager.Actor
		IClass declaringClass = cgNode.getMethod().getDeclaringClass();
		return actorSubclasses.contains(declaringClass);
	}

	@Override
	public boolean canTransform() {
		return true;
	}

	@Override
	protected int getMessageArgumentsIndex(MethodReference methodReference) {
		String methodSignature = methodReference.getSignature();
		if (methodSignature.endsWith(SEND_SIGNATURE_SUFFIX) || methodSignature.endsWith(CALL_SIGNATURE_SUFFIX) || methodSignature.endsWith(CREATE_SIGNATURE_SUFFIX_2)) {
			return 3;
		}
		if (methodSignature.endsWith(CREATE_SIGNATURE_SUFFIX_1)) {
			return 2;
		}
		return -1;
	}

	@Override
	protected void addMessageInvocationArguments(MessageInvocation messageInvocation, int messagePassingIndex, int passedPointerValueNumber) {
		CGNode cgNode = messageInvocation.getCGNode();
		SSAInstruction[] instructions = cgNode.getIR().getInstructions();
		int currentIndex = messagePassingIndex;
		while (true) {
			currentIndex--;
			SSAInstruction instruction = instructions[currentIndex];
			if (instruction != null) {
				if (instruction instanceof SSAArrayStoreInstruction && instruction.getUse(0) == passedPointerValueNumber) {
					int containedValueNumber = instruction.getUse(instruction.getNumberOfUses() - 1);
					PointerKey containedPointer = heapModel.getPointerKeyForLocal(cgNode, containedValueNumber);
					messageInvocation.addFirstArgument(currentIndex, computeTransitiveClosureOfPointedItems(containedPointer));
				} else if (instruction instanceof SSANewInstruction && instruction.getDef() == passedPointerValueNumber) {
					//Reached the definition of the serializable array, done.
					break;
				}
			}
		}
	}

}
