/**
 * Developed by: The Open Systems Lab
 *             University of Illinois at Urbana-Champaign
 *             Department of Computer Science
 *             Urbana, IL 61801
 *             http://osl.cs.uiuc.edu
 *
 * Contact: http://osl.cs.uiuc.edu/soter
 *
 * Copyright (c) 1998-2011
 * The University of Illinois Board of Trustees.
 *    All Rights Reserved.
 * 
 * Distributed under license: http://osl.cs.uiuc.edu/soter/LICENSE
 * 
 */
package edu.illinois.soter.transformation;

import java.io.FileOutputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import osl.util.DeepCopy;

import com.ibm.wala.classLoader.CallSiteReference;
import com.ibm.wala.classLoader.IClass;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.classLoader.ShrikeCTMethod;
import com.ibm.wala.classLoader.ShrikeClass;
import com.ibm.wala.ipa.callgraph.CGNode;
import com.ibm.wala.shrikeBT.CheckCastInstruction;
import com.ibm.wala.shrikeBT.MethodData;
import com.ibm.wala.shrikeBT.MethodEditor;
import com.ibm.wala.shrikeBT.Util;
import com.ibm.wala.shrikeBT.shrikeCT.ClassInstrumenter;
import com.ibm.wala.shrikeCT.ClassWriter;
import com.ibm.wala.shrikeCT.InvalidClassFileException;

import edu.illinois.soter.Debugger;
import edu.illinois.soter.messagedata.MessageArgument;
import edu.illinois.soter.messagedata.MessageInvocation;

/**
 * 
 * @author Stas Negara
 * 
 */
public class CodeTransformer {

	private final Map<IClass, ClassInstrumenter> classInstrumenters = new HashMap<IClass, ClassInstrumenter>();
	private final Map<IMethod, MethodEditor> methodEditors = new HashMap<IMethod, MethodEditor>();
	private final Map<MethodEditor, Set<Integer>> editedInstructions = new HashMap<MethodEditor, Set<Integer>>();

	public CodeTransformer() {

	}

	public void perform(String outputPath, Map<CGNode, Map<CallSiteReference, MessageInvocation>> messageInvocations) {
		for (Entry<CGNode, Map<CallSiteReference, MessageInvocation>> mapEntry : messageInvocations.entrySet()) {
			for (MessageInvocation messageInvocation : mapEntry.getValue().values()) {
				try {
					updateMethodCode(messageInvocation);
				} catch (Exception ex) {
					ex.printStackTrace();
				}
			}
		}
		applyCodeUpdates(outputPath);
	}

	//Update the code of the method that performs the message invocation such that arguments, 
	//which have to be passed by value, are deep copied.
	public void updateMethodCode(MessageInvocation messageInvocation) throws InvalidClassFileException {
		Iterator<MessageArgument> argumentsIterator = messageInvocation.getArgumentsIterator();
		while (argumentsIterator.hasNext()) {
			MessageArgument argument = argumentsIterator.next();
			if (argument.isPassByValue()) {
				MethodEditor methodEditor = getMethodEditor(messageInvocation.getCGNode());
				//Avoid inserting "deepCopy" call multiple times (this can happen when several CGNodes represent the same method)
				Set<Integer> editedSet = getEditedSet(methodEditor);
				if (!editedSet.contains(argument.getTransformationBytecodeIndex())) {
					editedSet.add(argument.getTransformationBytecodeIndex());
					methodEditor.insertBefore(argument.getTransformationBytecodeIndex(), new MethodEditor.Patch() {
						public void emitTo(MethodEditor.Output w) {
							w.emit(Util.makeInvoke(DeepCopy.class, "deepCopy"));
							w.emit(CheckCastInstruction.make("Ljava/io/Serializable;"));
						}
					});
				}
			}
		}
	}

	private Set<Integer> getEditedSet(MethodEditor methodEditor) {
		Set<Integer> editedSet = editedInstructions.get(methodEditor);
		if (editedSet == null) {
			editedSet = new HashSet<Integer>();
			editedInstructions.put(methodEditor, editedSet);
		}
		return editedSet;
	}

	private MethodEditor getMethodEditor(CGNode cgNode) throws InvalidClassFileException {
		ShrikeCTMethod method = (ShrikeCTMethod) cgNode.getMethod();
		MethodEditor methodEditor = methodEditors.get(method);
		if (methodEditor == null) {
			ShrikeClass declaringClass = (ShrikeClass) method.getDeclaringClass();
			ClassInstrumenter classInstrumenter = classInstrumenters.get(declaringClass);
			if (classInstrumenter == null) {
				classInstrumenter = new ClassInstrumenter(declaringClass.getReader());
				classInstrumenters.put(declaringClass, classInstrumenter);
			}
			MethodData methodData = classInstrumenter.visitMethod(method.getShrikeMethodIndex());
			methodEditor = new MethodEditor(methodData);
			methodEditor.beginPass();
			methodEditors.put(method, methodEditor);
		}
		return methodEditor;
	}

	private void applyCodeUpdates(String outputPath) {
		Iterator<MethodEditor> methodEditorsIterator = methodEditors.values().iterator();
		while (methodEditorsIterator.hasNext()) {
			MethodEditor methodEditor = methodEditorsIterator.next();
			methodEditor.applyPatches();
			methodEditor.endPass();
			Debugger.debugTransformedInstructions(methodEditor);
		}
		Iterator<ClassInstrumenter> classInstrumentersIterator = classInstrumenters.values().iterator();
		while (classInstrumentersIterator.hasNext()) {
			try {
				ClassInstrumenter classInstrumenter = classInstrumentersIterator.next();
				ClassWriter classWriter = classInstrumenter.emitClass();
				String fileName = outputPath + classInstrumenter.getReader().getName() + ".class";
				FileOutputStream outputStream = new FileOutputStream(fileName);
				outputStream.write(classWriter.makeBytes());
				outputStream.flush();
				outputStream.close();
			} catch (Exception ex) {
				ex.printStackTrace();
			}
		}
	}

}
